"""icl-hub REST API."""
from typing import Optional

import fastapi
import pydantic

from infractl.hub import config, jupyterhub

app = fastapi.FastAPI()


class JupyterHubEnableSshRequest(pydantic.BaseModel):
    """Request to enable SSH for JupyterHub session."""

    key: Optional[str] = pydantic.Field(
        title='SSH public key',
        default=None,
    )


@app.get('/healthz')
def get_health():
    """Health check."""
    return 'OK'


@app.get('/config')
def get_config_values():
    """Configuration values."""
    return config.get_dict()


@app.get('/config/{key}')
def get_config_value(key: str):
    """Configuration value."""
    return config.get().get(key)


@app.get('/jupyterhub/users')
def get_jupyterhub_users():
    """JupyterHub users."""
    return jupyterhub.list_users()


@app.post('/jupyterhub/ssh/enable/{user}')
def jupyterhub_ssh_enable(user: str, body: JupyterHubEnableSshRequest):
    """Enable SSH for JupyterHub session."""
    print(user, body.key)
    return jupyterhub.enable_ssh(user, body.key)


@app.post('/jupyterhub/ssh/disable/{user}')
def jupyterhub_ssh_disable(user: str):
    """Disable SSH for JupyterHub session."""
    jupyterhub.disable_ssh(user)
    return 'OK'



"""Wrapper to execute a Python program as a Prefect flow."""

import importlib
import pathlib
import runpy
import sys
from typing import Any, Dict, List, Optional, Union

import prefect


@prefect.flow(log_prints=True)
def wrap(
    program: str,
    entrypoint: Optional[str] = None,
    parameters: Union[Dict[str, Any], List[str], None] = None,
) -> Any:
    """Executes a Python program."""
    logger = prefect.get_run_logger()
    logger.info(f'Running {program=} {entrypoint=}) {parameters=}')
    if entrypoint:
        module = importlib.import_module(pathlib.Path(program).stem)
        target = getattr(module, entrypoint)
        kwargs = parameters or {}
        return target(**kwargs)
    else:
        if parameters:
            sys.argv = [program]
            sys.argv.extend(parameters)
        runpy.run_path(program, run_name='__main__')



"""Module for configuring logger."""

import datetime as dt
import logging
import os
import platform
from logging.handlers import RotatingFileHandler
from pathlib import Path

_LOGGER_CONFIGURED: dict = {}
_DEFAULT_NAMESPACE = 'x1'
# to share between different namespaces
_FILE_LOGGER = None
_ENV_LOGGING_TO_FILE = 'X1_LOGGING_TO_FILE'
_ENV_LOGGING_LEVEL = 'X1_LOGGING_LEVEL'


def _get_root_dir():
    # be careful when moving this function or file
    return Path(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))


def _create_logger(namespace: str, job_id: str, logfile_name: str) -> logging.Logger:
    """
    Create and configure logger as infractl expects it to be.

    Parameters
    ----------
    namespace : str
        Logging namespace to use, e.g. `_DEFAULT_NAMESPACE`.
    job_id : str
        Part of path to where logs are stored.
    logfile_name : str
        Name of the log file to create.

    Returns
    -------
    Logger
    """
    logger = logging.getLogger(namespace)
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    x1_logging_to_file = os.environ.get(_ENV_LOGGING_TO_FILE, None)
    if x1_logging_to_file not in ('TRUE', 'true', None):
        raise ValueError(
            f'env var: {_ENV_LOGGING_TO_FILE}={x1_logging_to_file}, '
            f'allowed values: {("TRUE", "true")}'
        )
    if x1_logging_to_file:
        global _FILE_LOGGER
        if _FILE_LOGGER is None:
            # Pathlib makes it OS agnostic.
            log_path = _get_root_dir() / 'logs'
            log_path.mkdir(exist_ok=True)

            # Add gitignore to the log directory.
            ignore_log_path = log_path / '.gitignore'
            if not ignore_log_path.exists():
                ignore_log_path.write_text('# Automatically generated by infractl.\n*\n')

            log_dir = log_path / f'job_{job_id}'
            log_dir.mkdir(parents=True, exist_ok=True)
            log_filename = log_dir / f'{logfile_name}.log'

            file_handler = RotatingFileHandler(
                filename=log_filename,
                mode='a',
                backupCount=10,
            )
            file_handler.setFormatter(formatter)
            file_handler.setLevel(logging.INFO)
            _FILE_LOGGER = file_handler
        else:
            file_handler = _FILE_LOGGER
        logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(os.environ.get(_ENV_LOGGING_LEVEL, 'WARNING'))
    logger.addHandler(stream_handler)

    return logger


def configure_logging(namespace) -> None:
    """Configure infractl logging by setting up directory structure and formatting."""
    current_timestamp = dt.datetime.now().strftime('%Y%m%d%H%M%S')

    root = logging.getLogger()
    # not sure why `PrefectConsoleHandler` is enabled by default as a root looger
    if len(root.handlers) == 1 and 'PrefectConsoleHandler' in str(root.handlers[0]):
        root.handlers = []

    logger = _create_logger(
        namespace,
        job_id=current_timestamp,
        logfile_name='info',
    )

    if namespace == _DEFAULT_NAMESPACE:
        logger.info('OS Version: %s', platform.platform())
        logger.info('Python Version:  %s', platform.python_version())


def get_logger(namespace: str = _DEFAULT_NAMESPACE) -> logging.Logger:
    """
    Configure infractl logger and returns the logger.

    Parameters
    ----------
    namespace : str, default: `_DEFAULT_NAMESPACE`
        Which namespace to use for logging.

    Returns
    -------
    logging.Logger
        The infractl logger.
    """
    if _LOGGER_CONFIGURED.get(namespace) is None:
        configure_logging(namespace)
        _LOGGER_CONFIGURED[namespace] = True

    return logging.getLogger(namespace)



"""ICL run (shortcut)"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import infractl.api.infrastructure
import infractl.api.runtime
import infractl.base
from infractl.api.deploy import deploy


async def run(
    program: infractl.base.Program,
    runtime: Optional[infractl.base.Runtime] = None,
    infrastructure: Optional[infractl.base.Infrastructure] = None,
    parameters: Union[Dict[str, Any], List[str], None] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> infractl.base.ProgramRun:
    """Deploys and runs a program by accepting a union of parameters
        from infractl.api.deploy() and infractl.base.program.run()

    Args:
        program: a program to deploy, can be a file name, ICL program or Prefect Flow instance.
        runtime: an optional program runtime to use for deployment.
        infrastructure: an optional infrastructure to use for deployment.
        timeout: an optional timeout to use for deployment. This value is the same
                 for deploy and run functions.
        kwargs: other parameters for deployment.
    """
    deployed_program = await deploy(program, runtime, infrastructure, timeout, **kwargs)

    # Run the deployed program
    return await deployed_program.run(parameters, timeout)



"""Infrastructure for ICL cluster."""

import functools
from typing import Optional, Tuple

import dynaconf

import infractl.base


class IclInfrastructureImplementation(
    infractl.base.InfrastructureImplementation,
    registration_name='ICL',
):
    """ICL Infrastructure specification."""

    # Custom settings to use instead of global infractl.base.SETTINGS
    _settings: Optional[dynaconf.Dynaconf] = None

    @functools.cached_property
    def address(self) -> str:
        """Gets infrastructure address."""
        target = self.infrastructure
        if target.address:
            return target.address
        current_settings = self._settings or infractl.base.SETTINGS
        return current_settings.get('default_address', 'localtest.me')

    @functools.cached_property
    def gpus(self) -> Optional[Tuple[str, int]]:
        """Gets GPUs limits."""
        target = self.infrastructure
        if target.gpus:
            return target.gpus
        current_settings = self._settings or infractl.base.SETTINGS
        return current_settings.get('default_gpus_limits', None)



"""ICL base."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infractl.base.infrastructure import (
    Infrastructure,
    InfrastructureImplementation,
    get_infrastructure_implementation,
)
from infractl.base.program import DeployedProgram, Program, ProgramRun, Runnable
from infractl.base.registry import RegisteredClass
from infractl.base.runtime import (
    Runtime,
    RuntimeDependencies,
    RuntimeFile,
    RuntimeImplementation,
    get_runtime_implementation,
)

if TYPE_CHECKING:
    import dynaconf  # pylint: disable=ungrouped-imports

SETTINGS: dynaconf.Dynaconf


def _settings() -> dynaconf.Dynaconf:
    """Loads settings from the nested package to avoid import-time initialization."""
    # pylint: disable=import-outside-toplevel
    import infractl.base.settings as base_settings

    globals()['SETTINGS'] = base_settings.SETTINGS
    return base_settings.SETTINGS


# pylint: disable=invalid-name
def __getattr__(name) -> Any:
    if name == 'SETTINGS':
        return _settings()
    # Implicit else
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')


__all__ = [
    'DeployedProgram',
    'Infrastructure',
    'InfrastructureImplementation',
    'get_infrastructure_implementation',
    'Program',
    'ProgramRun',
    'RegisteredClass',
    'Runnable',
    'Runtime',
    'RuntimeDependencies',
    'RuntimeFile',
    'RuntimeImplementation',
    'get_runtime_implementation',
    'SETTINGS',
]



"""ICL infrastructure."""

import infractl.base


def default_infrastructure() -> infractl.base.Infrastructure:
    """Returns a default ICL infrastructure."""
    return infrastructure(kind='ICL')


def infrastructure(*args, kind: str = 'ICL', **kwargs) -> infractl.base.Infrastructure:
    if kind == 'ICL':
        # import required plugin
        # pylint: disable=import-outside-toplevel, unused-import
        from infractl.plugins import icl_infrastructure  # noqa

        return infractl.base.Infrastructure(*args, kind=kind, **kwargs)

    raise NotImplementedError(f'Infrastructure {kind} is not implemented')



"""ICL Python API."""


from infractl.api.deploy import deploy
from infractl.api.infrastructure import infrastructure
from infractl.api.program import program
from infractl.api.run import run
from infractl.api.runtime import runtime

__all__ = [
    'infrastructure',
    'runtime',
    'program',
    'deploy',
    'run',
]



"""ICL Prefect runtime implementation.

Environment variable PREFECT_API_URL needs to be set to the URL of a Prefect server running in the
cluster.

"""

import asyncio
import contextlib
import copy
import functools
import os
import pathlib
import sys
import tarfile
import tempfile
import urllib.parse
from typing import Any, Callable, Dict, List, Optional, Union

import dynaconf
import prefect.blocks.core as blocks
import pydantic
from prefect import deployments, filesystems, infrastructure, settings
from prefect.client import orchestration
from prefect.server.api import server

import infractl
import infractl.base
import infractl.plugins.prefect_runtime.utils as prefect_utils
from infractl.logging import get_logger
from infractl.plugins import icl_infrastructure, prefect_runtime

logger = get_logger()


DEFAULT_ITEMS_TO_IGNORE = [
    # Prefect artifacts
    '.prefectignore',
    # Python artifacts
    '__pycache__/',
    '*.py[cod]',
    '*$py.class',
    '*.egg-info/',
    '*.egg',
    # Type checking artifacts
    '.mypy_cache/',
    '.dmypy.json',
    'dmypy.json',
    '.pyre/',
    # IPython
    'profile_default/',
    'ipython_config.py',
    '*.ipynb_checkpoints/*',
    # Environments
    '.python-version',
    '.env',
    '.venv',
    'env/',
    'venv/',
    '.conda/',
    # MacOS
    '.DS_Store',
    # Dask
    'dask-worker-space/',
    # Editors
    '.idea/',
    '.vscode/',
]


ManifestFilter = Callable[[infrastructure.KubernetesManifest], infrastructure.KubernetesManifest]
"""Accepts Kubernetes manifest as a dictionary and returns an updated manifest."""


class PrefectRuntimeError(Exception):
    """Prefect runtime error."""


def strip_file_scheme(uri: str) -> str:
    """Strips "file" schema from the URI.

    Examples:
        file:///C:/Windows -> C:/Windows
        file:///home -> /home
        file:local -> local
    """
    path = urllib.parse.unquote(urllib.parse.urlparse(uri).path)
    if sys.platform == 'win32' and path.startswith('/'):
        # workaround to remove leading slash on Windows
        return path[1:]
    return path


def stable_identity() -> str:
    """Generates stable identity for the current user.

    Uses environment variables `JUPYTERHUB_USER`, `USER` and sanitizes the value to make it
    compatible with Prefect names.
    """
    identity = os.environ.get('JUPYTERHUB_USER') or os.environ.get('USER') or 'unknown'
    return prefect_runtime.sanitize(identity)


def upload_files(files: List[infractl.base.RuntimeFile], target_path: pathlib.Path):
    """Uploads files to the specified directory."""
    # The specified directory has the following structure:
    # cwd.tar  - tarball to extract to the current directory in runtime.
    with tarfile.open(name=target_path / 'cwd.tar', mode='w') as cwd:
        for file in files:
            if file.src.endswith('/'):
                src = pathlib.Path(file.src)
                dst = ''
                if file.dst:
                    # dst is expected to be a directory, adding a trailing / if missing
                    dst = file.dst if file.dst.endswith('/') else f'{file.dst}/'
                for path in src.rglob('*'):
                    relative_path = path.relative_to(src)
                    cwd.add(name=path, arcname=f'{dst}{relative_path}')
            else:
                dst = file.dst or file.src
                cwd.add(name=file.src, arcname=pathlib.Path(dst).name)


class PrefectBlock(pydantic.BaseModel):
    """Prefect Block specification."""

    kind: str
    name: str
    block: Optional[blocks.Block] = None

    @property
    def full_name(self):
        return f'{self.kind}/{self.name}'

    async def save(self, **kwargs):
        return await self.block.save(self.name, **kwargs)


class PrefectProgramRunner(infractl.base.Runnable):
    """Prefect program."""

    def __init__(
        self,
        prefect_client: orchestration.PrefectClient,
        prefect_deployment: deployments.Deployment,
    ):
        self.prefect_client = prefect_client
        self.prefect_deployment = prefect_deployment

    @property
    def deployment(self) -> deployments.Deployment:
        """Returns a Prefect deployment for this program."""
        return self.prefect_deployment

    async def cancel(self, flow_id):
        """Cancel a flow run by ID."""
        await prefect_utils.cancel(self.prefect_client, flow_id)

    async def run(
        self,
        parameters: Union[Dict[str, Any], List[str], None] = None,
        timeout: Optional[float] = None,
        detach: bool = False,
    ) -> prefect_runtime.PrefectProgramRun:
        """Runs flow synchronously.

        Note that currently Prefect does not support returning the result.
        """

        # setting detach to True is the same as setting timeout to 0
        if detach:
            timeout = 0

        deployment_name = f'{self.prefect_deployment.flow_name}/{self.prefect_deployment.name}'
        logger.info('Running deployment "%s" with timeout="%s" seconds', deployment_name, timeout)
        flow_run = await deployments.run_deployment(
            name=deployment_name,
            parameters=parameters,
            client=self.prefect_client,
            timeout=timeout,
        )
        if not flow_run.state.is_final() and timeout not in (None, 0):
            logger.info('Cancel flow run "%s" by timeout="%s" seconds', flow_run.name, timeout)
            await self.cancel(flow_run.id)
            raise asyncio.exceptions.TimeoutError
        logger.info('FlowRun %s %s', flow_run.name, flow_run.state)
        return prefect_runtime.PrefectProgramRun(flow_run, self.prefect_client)


class PythonProgramRunner(PrefectProgramRunner):
    """Runs a Python program.

    TODO: potentially use types.MethodType to override `run` for the existing runner instance?
    """

    program: str
    entrypoint: Optional[str]
    runner: PrefectProgramRunner

    def __init__(
        self,
        runner: PrefectProgramRunner,
        program: str,
        entrypoint: Optional[str],
    ):
        super().__init__(runner.prefect_client, runner.prefect_deployment)
        self.runner = runner
        self.program = program
        self.entrypoint = entrypoint

    async def cancel(self, flow_id):
        """Cancels a flow run by ID."""
        await self.runner.cancel(flow_id)

    async def run(
        self,
        parameters: Union[Dict[str, Any], List[str], None] = None,
        timeout: Optional[float] = None,
        detach: bool = False,
    ) -> prefect_runtime.PrefectProgramRun:
        """Runs flow."""
        if self.entrypoint:
            if parameters and not isinstance(parameters, Dict):
                raise ValueError('For a Python function parameters must be a dict')
        else:
            if parameters and not isinstance(parameters, List):
                raise ValueError('For a Python program parameters must be a list')

        flow_parameters = {
            'program': self.program,
            'entrypoint': self.entrypoint,
            'parameters': parameters,
        }
        return await self.runner.run(parameters=flow_parameters, timeout=timeout, detach=detach)


class PrefectRuntimeImplementation(
    infractl.base.RuntimeImplementation, registration_name='prefect'
):
    """Prefect runtime implementation."""

    # Custom settings to use instead of global infractl.base.SETTINGS
    _settings: Optional[dynaconf.Dynaconf] = None
    _files_block: str = ''
    _script: str = ''

    def __init__(
        self,
        runtime: Optional[infractl.base.runtime.Runtime] = None,
        infrastructure_implementation: Optional[infractl.base.InfrastructureImplementation] = None,
    ):
        if not isinstance(
            infrastructure_implementation, icl_infrastructure.IclInfrastructureImplementation
        ):
            raise NotImplementedError(
                f'{infrastructure_implementation.__class__.__name__} is not supported'
            )
        super().__init__(
            runtime=runtime, infrastructure_implementation=infrastructure_implementation
        )

    @functools.cached_property
    def prefect_client(self):
        """Returns Prefect client."""
        if self.prefect_api_url:
            return orchestration.PrefectClient(self.prefect_api_url)
        return orchestration.PrefectClient(server.create_app(ephemeral=True))

    @functools.cached_property
    def prefect_api_url(self) -> Optional[str]:
        """Returns Prefect API URL.

        If the infrastructure address is not set (empty string) or 'local' then this method returns
        None, which means that an ephemeral instance of Prefect will be used.
        """
        # TODO: support different schemas
        if (
            self.infrastructure_implementation.address
            and self.infrastructure_implementation.address != 'local'
        ):
            return f'http://prefect.{self.infrastructure_implementation.address}/api'
        return None

    async def deploy(
        self,
        program: prefect_runtime.PrefectProgram,
        customizations: Optional[List[Dict[str, Any]]] = None,
        manifest_filter: Optional[ManifestFilter] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        """Deploys a program."""
        with settings.temporary_settings(updates={settings.PREFECT_API_URL: self.prefect_api_url}):
            if isinstance(program, prefect_runtime.PythonProgram):
                return await self.deploy_python_program(
                    program,
                    customizations=customizations,
                    manifest_filter=manifest_filter,
                    name=name,
                    **kwargs,
                )
            else:
                return await self.deploy_prefect_program(
                    program,
                    customizations=customizations,
                    manifest_filter=manifest_filter,
                    name=name,
                    **kwargs,
                )

    async def deploy_python_program(
        self,
        program: prefect_runtime.PythonProgram,
        customizations: Optional[List[Dict[str, Any]]] = None,
        manifest_filter: Optional[ManifestFilter] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> infractl.base.DeployedProgram:
        """Deploys a Python program wrapped in a Prefect flow."""
        # Running a Prefect wrapper for a Python program requires a custom runtime and a custom
        # runner.

        self.runtime = copy.copy(self.runtime)
        self.runtime.files = copy.copy(self.runtime.files)
        self.runtime.files.extend(program.files)

        deployed_program = await self.deploy_prefect_program(
            program,
            customizations=customizations,
            manifest_filter=manifest_filter,
            name=name,
            **kwargs,
        )
        deployed_program.runner = PythonProgramRunner(
            runner=deployed_program.runner,
            program=program.program,
            entrypoint=program.entrypoint,
        )

        return deployed_program

    async def deploy_prefect_program(
        self,
        program: prefect_runtime.PrefectProgram,
        customizations: Optional[List[Dict[str, Any]]] = None,
        manifest_filter: Optional[ManifestFilter] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> infractl.base.DeployedProgram:
        """Deploys a Prefect program."""
        if not isinstance(program, prefect_runtime.PrefectProgram):
            raise NotImplementedError(f'PrefectProgram expected, got {program.__class__.__name__}')

        flow_file_name = pathlib.Path(program.path).name
        prefect_flow = program.flow
        # use a custom flow name if specified
        if name:
            # create a copy of the flow object with a new name to make sure to keep to original
            # object intact for the case it was imported as Python object.
            prefect_flow = prefect_flow.with_options(name=name)

        prefect_storage_block = await self.create_code_block(prefect_flow.name)
        if self.runtime.files:
            await self.create_files_block(prefect_flow.name)
        prefect_infrastructure_block = await self.create_infrastructure_block(
            prefect_flow.name,
            customizations=customizations,
            manifest_filter=manifest_filter,
        )

        prefect_flow_entrypoint = f'{flow_file_name}:{prefect_flow.fn.__name__}'
        prefect_deployment = await deployments.Deployment.build_from_flow(
            flow=prefect_flow,
            # Note that Prefect shows deployments as "{flow_name}/{deployment_name}".
            name=program.deployment_name,
            storage=await blocks.Block.load(
                prefect_storage_block.full_name,
                client=self.prefect_client,
            ),
            infrastructure=await blocks.Block.load(
                prefect_infrastructure_block.full_name,
                client=self.prefect_client,
            ),
            work_queue_name=self.settings('prefect_queue', 'prod'),
            apply=False,
            skip_upload=True,
            entrypoint=prefect_flow_entrypoint,
        )

        deployment_keys = prefect_deployment.dict().keys()
        deployment_dict = {}
        for key, value in kwargs.items():
            if key in deployment_keys:
                deployment_dict[key] = value
        if deployment_dict:
            await prefect_deployment.update(**deployment_dict)

        cwd = os.getcwd()
        os.chdir(pathlib.Path(program.path).absolute().parent)
        ignore_file_name: Optional[str] = None
        try:
            with tempfile.NamedTemporaryFile(delete=False) as ignore_file:
                ignore_file_name = ignore_file.name
                ignore_file.write('\n'.join(DEFAULT_ITEMS_TO_IGNORE).encode('utf-8'))
                ignore_file.flush()
                ignore_file.close()
                await prefect_deployment.upload_to_storage(
                    prefect_storage_block.full_name,
                    ignore_file=ignore_file_name,
                )
        finally:
            os.chdir(cwd)
            if ignore_file_name:
                with contextlib.suppress(FileNotFoundError):
                    os.remove(ignore_file_name)

        await prefect_deployment.apply(upload=False)

        return infractl.base.DeployedProgram(
            program=program,
            runner=PrefectProgramRunner(self.prefect_client, prefect_deployment),
        )

    @functools.cached_property
    def remote_storage_settings(self):
        """Gets storage settings for Prefect RemoteFileSystem."""
        storage_settings = self.settings('prefect_storage_fsspec', {})
        storage_settings.setdefault('key', 'x1miniouser')
        storage_settings.setdefault('secret', 'x1miniopass')
        storage_settings.setdefault('use_ssl', False)
        client_kwargs = storage_settings.setdefault('client_kwargs', {})
        if 'endpoint_url' not in client_kwargs:
            # TODO: support different schemas
            client_kwargs[
                'endpoint_url'
            ] = f'http://s3.{self.infrastructure_implementation.address}'
        return storage_settings

    def define_storage_block(self, base_path: str, name: str) -> PrefectBlock:
        """Defines Prefect storage block.

        The following Prefect storage blocks are supported:
        * LocalFileSystem (scheme is "file")
        * RemoteFileSystem (scheme is not "file")
        """
        if base_path.startswith('file:'):
            block = filesystems.LocalFileSystem(
                # Note the trailing slash, https://github.com/PrefectHQ/prefect/issues/8710
                basepath=f'{strip_file_scheme(base_path)}/{name}/',
            )
            return PrefectBlock(kind='local-file-system', name=name, block=block)
        else:
            block = filesystems.RemoteFileSystem(
                # Note the trailing slash, https://github.com/PrefectHQ/prefect/issues/8710
                basepath=f'{base_path}/{name}/',
                settings=self.remote_storage_settings,
            )
            return PrefectBlock(kind='remote-file-system', name=name, block=block)

    async def create_code_block(self, flow_name: str) -> PrefectBlock:
        """Creates and saves Prefect storage block."""
        base_path = self.settings('prefect_storage_basepath', 's3://prefect')
        block = self.define_storage_block(base_path, flow_name)
        await block.save(overwrite=True, client=self.prefect_client)
        return block

    async def create_files_block(self, flow_name: str):
        """Creates a Prefect storage block, upload files and script for infractl.prefect.engine."""
        identity = stable_identity()
        storage_path = self.settings('prefect_storage_basepath', 's3://prefect')
        base_path = f'{storage_path}/_files/{identity}'
        block = self.define_storage_block(base_path, f'{identity}-{flow_name}-files')
        await block.save(overwrite=True, client=self.prefect_client)
        await self.upload_files(block)

    async def upload_files(self, block: PrefectBlock):
        """Uploads files to a files block."""
        # Use a unique file name to make sure it does not overlap with user's files.
        # This script is located in the temporary directory in runtime, but executed from the
        # current directory in runtime. So "$PWD" (or ".") points to the current directory,
        # "$SCRIPT_PATH" points to the temporary directory with this script and all required files.
        self._script = '81503f92-f80a-4a8f-855c-b399f2ec41df.sh'
        script_lines = [
            '#!/bin/bash',
            'set -e',
            'SCRIPT_PATH=$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd )',
            'if [[ -f "$SCRIPT_PATH/cwd.tar" ]]; then tar xvf "$SCRIPT_PATH/cwd.tar"; fi',
        ]
        # Create a temporary directory with the required files, generate bash script to extract the
        # files to the corresponding locations in runtime, upload the whole directory content to the
        # file storage.
        with tempfile.TemporaryDirectory() as dirname:
            target_path = pathlib.Path(dirname)
            script_path = target_path / self._script
            upload_files(self.runtime.files, target_path)
            script_path.write_text('\n'.join(script_lines))
            await block.block.put_directory(local_path=dirname)
        self._files_block = block.full_name

    async def create_infrastructure_block(
        self,
        flow_name: str,
        customizations: Optional[List[Dict[str, Any]]] = None,
        manifest_filter: Optional[ManifestFilter] = None,
    ) -> PrefectBlock:
        """Creates Prefect infrastructure block."""
        block = self.kubernetes_job(customizations=customizations, manifest_filter=manifest_filter)
        await block.save(flow_name, overwrite=True, client=self.prefect_client)
        return PrefectBlock(kind='kubernetes-job', name=flow_name)

    def kubernetes_job(
        self,
        customizations: Optional[List[Dict[str, Any]]] = None,
        manifest_filter: Optional[ManifestFilter] = None,
    ) -> infrastructure.KubernetesJob:
        """Creates Prefect KubernetesJob block."""
        job_args = {
            'customizations': customizations.copy() if customizations else [],
            'env': self.runtime.environment.copy() if self.runtime.environment else {},
        }

        # TODO: move the default image to discover
        image = self.settings('prefect_image', 'pbchekin/x1-prefect:2.13.0-python3.9-20231010')
        if image:
            job_args['image'] = image

        # TODO: support different targets runtime dependencies, such as using a custom docker image
        if self.runtime.dependencies.pip:
            # TODO: check if environment variable EXTRA_PIP_PACKAGES is set and merge if necessary
            job_args['env']['EXTRA_PIP_PACKAGES'] = ' '.join(self.runtime.dependencies.pip)

        # TODO: support different file injection methods, such as using a custom docker image
        if self.runtime.files:
            job_args['command'] = [
                'python',
                '-m',
                'x1.prefect.engine',
                '--block',
                self._files_block,
                '--script',
                self._script,
            ]

        # name is required for `build_job`
        job = infrastructure.KubernetesJob(name='foo', **job_args)
        manifest = job.build_job()

        manifest = self.manifest_filter(manifest)
        if manifest_filter:
            manifest = manifest_filter(manifest)

        metadata = manifest['metadata']
        metadata.pop('namespace', None)
        metadata.pop('generateName', None)

        # The following parameters are applied already in manifest, do not set them again
        job_args.pop('customizations', None)
        job_args.pop('env')

        return infrastructure.KubernetesJob(job=manifest, **job_args)

    def manifest_filter(
        self,
        manifest: infrastructure.KubernetesManifest,
    ) -> infrastructure.KubernetesManifest:
        """Filters Kubernetes Job manifest.

        Adds shared volume, if enabled.
        Adds GPU resource, if enabled
        """
        manifest = copy.deepcopy(manifest)
        prefect_pod = manifest['spec']['template']['spec']
        prefect_container = prefect_pod['containers'][0]

        # TODO: make it customizable, also use discover
        shared_volume_mount = self.settings('prefect_shared_volume_mount')
        if shared_volume_mount:
            volumes = prefect_pod.setdefault('volumes', [])
            volumes.append(
                {
                    'name': 'shared-volume',
                    'persistentVolumeClaim': {
                        'claimName': 'shared-volume',
                    },
                },
            )
            volume_mounts = prefect_container.setdefault('volumeMounts', [])
            volume_mounts.append(
                {
                    'name': 'shared-volume',
                    'mountPath': shared_volume_mount,
                },
            )

        gpus = self.infrastructure_implementation.gpus
        if gpus:
            resources = prefect_container.setdefault('resources', {})
            limits = resources.setdefault('limits', {})
            limits[gpus[0]] = str(gpus[1])

        return manifest

    def settings(self, name: str, default_value: Any = None) -> Any:
        """Return a setting value for this infrastructure address."""
        current_settings = self._settings or infractl.base.SETTINGS
        return current_settings.get(
            f'{self.infrastructure_implementation.address}.{name}', default_value
        )



"""Prefect flow to build and push Docker image."""

import io
import os
import re
import signal
import subprocess
import time
from typing import Any, Dict

import prefect


class BuildError(Exception):
    """Build error."""


def wait_docker(timeout=600):
    """Waits for Docker daemon.

    Since we run dind in a separate container it takes some time for dind to start and generate
    client certificates, so there is a retry loop.
    """
    # import locally to avoid dependency on psutil when deploying this flow.
    # pylint: disable=import-outside-toplevel
    import psutil

    logger = prefect.get_run_logger()
    logger.info('Waiting for dockerd')
    for retry in range(timeout):
        dockerd_pid = [
            process.pid for process in psutil.process_iter() if process.name() == 'dockerd'
        ]
        if dockerd_pid:
            logger.info('dockerd is running')
            break
        time.sleep(1)
    else:
        raise BuildError(f'dockerd is not running after {timeout} seconds, giving up')

    for _ in range(max(timeout - retry, 1)):
        # pylint: disable=subprocess-run-check
        process = subprocess.run(['docker', 'info'], capture_output=True)
        if process.returncode == 0:
            logger.info('dockerd is responding')
            return
        time.sleep(1)

    raise BuildError(f'dockerd is not responding after {timeout} seconds, giving up')


def kill_docker(timeout=30):
    """Kills Docker daemon.

    Docker daemon is running in a separate container, but in the same pod. This pod has
    `shareProcessNamespace` set to true, so we can see the processes from other container and send
    a signal to terminate.

    Args:
        timeout: timeout in seconds to wait for docker termination.
    """
    # import locally to avoid dependency on psutil when deploying this flow.
    # pylint: disable=import-outside-toplevel
    import psutil

    logger = prefect.get_run_logger()
    docker_init_pid = [
        process.pid for process in psutil.process_iter() if process.name() == 'docker-init'
    ]
    if not docker_init_pid:
        logger.info('docker-init not found, nothing to do')
        return

    logger.info('Terminating docker-init and waiting for docker processes to finish')
    os.kill(docker_init_pid[0], signal.SIGTERM)
    docker_processes = []
    for _ in range(timeout):
        docker_processes = [
            process.name()
            for process in psutil.process_iter()
            if process.name() in ('docker-init', 'dockerd', 'containerd')
        ]
        if not docker_processes:
            logger.info('Docker has been terminated, exiting')
            return
        time.sleep(1)
    logger.info(f'Giving up after waiting {timeout}s for docker processes to finish, exiting')
    logger.info(f'Remaining processes: {", ".join(docker_processes)}')


@prefect.flow(log_prints=True)
def build(build_args: Dict[str, Any], registry: str):
    """Builds and pushes Docker image."""
    logger = prefect.get_run_logger()
    logger.info(f'{build_args=}')

    tag = build_args['tag']
    registry = re.sub(r'^https?://', '', registry)
    full_tag = f'{registry}/{tag}'

    try:
        wait_docker()

        with subprocess.Popen(
            ['docker', 'build', '--tag', full_tag, 'context'],
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        ) as process:
            for line in io.TextIOWrapper(process.stdout, encoding='utf-8'):
                print(line.rstrip())
            process.wait()
            if process.returncode != 0:
                raise BuildError('Build failed')

        with subprocess.Popen(
            ['docker', 'push', full_tag],
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        ) as process:
            for line in io.TextIOWrapper(process.stdout, encoding='utf-8'):
                print(line.rstrip())
            process.wait()
            if process.returncode != 0:
                raise BuildError('Push failed')
    finally:
        kill_docker()



"""icl-hub server."""

from infractl.hub import root


@root.cli.group()
def server():
    """Manage icl-hub server."""


@server.command()
def start():
    """Start icl-hub server."""
    # pylint: disable=import-outside-toplevel
    import uvicorn

    uvicorn.run('infractl.hub.rest:app', host='0.0.0.0', port=8000, reload=False)



"""Builds and pushes a Docker image to Docker registry using a local Docker daemon."""
import itertools
import re
import tempfile
from typing import Optional

import docker.errors
import python_docker.base
import python_docker.registry

import infractl.docker
import infractl.logging

logger = infractl.logging.get_logger(__name__)


class Builder(infractl.docker.Builder):
    """Builds and pushes a Docker image to Docker registry using a local Docker daemon."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(
        self,
        stream_callback: Optional[infractl.docker.StreamCallback] = infractl.docker.stdout_callback,
        **kwargs,
    ) -> infractl.docker.Image:
        """Builds and pushes a Docker image to Docker registry using a local Docker daemon.

        Arguments:
        https://docker-py.readthedocs.io/en/stable/images.html#docker.models.images.ImageCollection.build
        """

        # TODO: image full tag (name:tag) is currently required!
        for key in ('path', 'tag'):
            if key not in kwargs:
                raise ValueError(f'Argument {key} is required')

        logger.info(
            'Building image %s from %s', kwargs.get('tag', '(untagged)'), kwargs.get('path', 'N/A')
        )

        client = docker.from_env()
        client_kwargs = kwargs.copy()
        # The returned stream will be decoded into dicts on the fly
        client_kwargs['decode'] = True
        # TODO: filter valid arguments for client.api.build
        resp = client.api.build(**client_kwargs)

        if isinstance(resp, str):
            return infractl.docker.Image.from_full_name(full_name=resp)

        last_event = None
        image_id = None
        result_stream, stream = itertools.tee(resp)
        for chunk in stream:
            if stream_callback and 'stream' in chunk:
                stream_callback(chunk['stream'].strip())
            if 'error' in chunk:
                raise docker.errors.BuildError(chunk['error'], result_stream)
            if not image_id and 'stream' in chunk:
                match = re.search(r'(^Successfully built |sha256:)([0-9a-f]+)$', chunk['stream'])
                if match:
                    image_id = match.group(2)
            last_event = chunk

        if image_id:
            image = infractl.docker.Image.from_full_name(id=image_id, full_name=kwargs.get('tag'))
            self.push(image)
            return image

        raise docker.errors.BuildError(last_event or 'Unknown', result_stream)

    def push(self, image: infractl.docker.Image):
        """Pushes the existing Docker image to registry."""
        image_id = image.full_name
        client = docker.from_env()
        docker_image = client.images.get(image_id)
        logger.info('Saving image %s', image_id)
        with tempfile.TemporaryDirectory() as tmp_dir:
            tar_file_name = f'{tmp_dir}/{image_id}.tar'
            with open(tar_file_name, 'wb') as tar_file:
                for chunk in docker_image.save(named=True):
                    tar_file.write(chunk)

            pd_images = python_docker.base.Image.from_filename(tar_file_name)
            logger.info('Pushing image %s to %s', image_id, self.registry_endpoint)
            pd_registry = python_docker.registry.Registry(hostname=self.registry_endpoint)
            for pd_image in pd_images:
                pd_registry.push_image(pd_image)



"""Kubernetes utils."""

from __future__ import annotations

import os
from typing import Optional

from kubernetes import client, config

DEFAULT_KUBE_API: Optional[KubeApi] = None


class KubeApi:
    """Wrapper for Kubernetes client.

    TODO: handle caching and re-authentication.
    """

    def __init__(self):
        if 'KUBERNETES_SERVICE_HOST' in os.environ and 'KUBERNETES_SERVICE_PORT' in os.environ:
            config.load_incluster_config()
            self.configuration = client.Configuration.get_default_copy()
        else:
            self.configuration = client.Configuration()
            config.load_kube_config(client_configuration=self.configuration)
        self.api_client = client.ApiClient(configuration=self.configuration)

    def core_v1(self) -> client.CoreV1Api:
        return client.CoreV1Api(api_client=self.api_client)


def api() -> KubeApi:
    """Returns default KubeApi object."""
    global DEFAULT_KUBE_API
    if DEFAULT_KUBE_API is None:
        DEFAULT_KUBE_API = KubeApi()
    return DEFAULT_KUBE_API



"""ICL Prefect program management."""

from __future__ import annotations

import importlib
import logging
import pathlib
import sys
import warnings
from typing import Any, Dict, List, Optional, Union

import anyio
import prefect
from prefect.client import orchestration
from prefect.client.schemas.filters import LogFilter
from prefect.utilities import importtools

import infractl.base
import infractl.plugins.prefect_runtime.utils as prefect_utils
from infractl.logging import get_logger
from infractl.plugins import prefect_runtime

logger = get_logger()


class FlowError(Exception):
    """Flow error."""


def load_flows(path: str) -> Dict[str, prefect.Flow]:
    """Returns a dictionary of Prefect Flow objects found in the specified path."""
    module = importtools.load_script_as_module(path)
    flows: Dict[str, Any] = {}
    for name in dir(module):
        value = getattr(module, name)
        if isinstance(value, prefect.Flow):
            flows[name] = value
    return flows


class PrefectProgramRun(infractl.base.ProgramRun):
    """Class for checking status and getting results."""

    _flow_run: prefect.client.schemas.FlowRun = None

    def __init__(
        self, flow_run: prefect.client.schemas.FlowRun, prefect_client: orchestration.PrefectClient
    ):
        self._flow_run = flow_run
        self._prefect_client = prefect_client

    def get_flow_run(self) -> prefect.client.schemas.FlowRun:
        return self._flow_run

    async def update(self) -> None:
        self._flow_run = await self._prefect_client.read_flow_run(self._flow_run.id)

    async def wait(self, poll_interval=5) -> None:
        """Wait until the terminal (COMPLETED, CANCELLED, FAILED, CRASHED) state is reached."""
        while True:
            await self.update()
            flow_run = self.get_flow_run()
            if flow_run.state.is_final():
                return
            await anyio.sleep(poll_interval)

    def is_scheduled(self) -> bool:
        return self.get_flow_run().state.is_scheduled()

    def is_pending(self) -> bool:
        return self.get_flow_run().state.is_pending()

    def is_running(self) -> bool:
        return self.get_flow_run().state.is_running()

    def is_completed(self) -> bool:
        return self.get_flow_run().state.is_completed()

    def is_failed(self) -> bool:
        return self.get_flow_run().state.is_failed()

    def is_crashed(self) -> bool:
        return self.get_flow_run().state.is_crashed()

    def is_cancelling(self) -> bool:
        return self.get_flow_run().state.is_running()

    def is_cancelled(self) -> bool:
        return self.get_flow_run().state.is_cancelled()

    def is_final(self) -> bool:
        return self.get_flow_run().state.is_final()

    def is_paused(self) -> bool:
        return self.get_flow_run().state.is_paused()

    def __repr__(self) -> str:
        """Returns a string representation.

        Note that JupyterLab uses __repr__ instead of __str__.
        """
        flow_run = self.get_flow_run()
        return f'{flow_run.name} ({flow_run.state.name})'

    async def logs(self) -> list[str]:
        """View logs for a flow run by ID."""
        # Copied (with modification) from:
        #  @flow_run_app.command()
        #  async def logs(id: UUID, ...):

        flow_run = self.get_flow_run()
        flow_run_id = flow_run.id
        log_filter = LogFilter(flow_run_id={'any_': [flow_run_id]})
        logger.info('Created log filter: %s', log_filter)

        # Get logs by using filter
        page_logs = await self._prefect_client.read_logs(
            log_filter=log_filter,
        )
        logger.info('Received %s logs entries with filter', len(page_logs))

        logs: list[str] = [None] * len(page_logs)
        for idx, log in enumerate(page_logs):
            logs[idx] = (
                f'{log.timestamp} |'
                f' {logging.getLevelName(log.level):7s} | Flow run'
                f' {flow_run.name!r} - {log.message}'
            )
        return logs

    async def stream_logs(self, poll_interval=5, file=None) -> None:
        """
        Stream logs until the terminal (COMPLETED, CANCELLED, FAILED, CRASHED) state is reached.
        """
        timestamp = None
        flow_run = self.get_flow_run()
        flow_run_id = flow_run.id

        while True:
            await self.update()
            flow_run = self.get_flow_run()

            log_filter = LogFilter(
                flow_run_id={'any_': [flow_run_id]}, timestamp={'after_': timestamp}
            )
            logger.info('Created log filter: %s', log_filter)

            # Get logs by using filter
            page_logs = await self._prefect_client.read_logs(
                log_filter=log_filter,
            )
            logger.info('Received %s logs entries with filter', len(page_logs))

            if timestamp and len(page_logs):
                # since `'after_': timestamp` boundary is inclusive we need to drop
                # first log entry to avoid duplication
                page_logs.pop(0)
            logs: list[str] = [None] * len(page_logs)
            for idx, log in enumerate(page_logs):
                logs[idx] = (
                    f'{log.timestamp} |'
                    f' {logging.getLevelName(log.level):7s} | Flow run'
                    f' {flow_run.name!r} - {log.message}'
                )

            for log in logs:
                print(log, file=sys.stdout if file is None else file)

            if flow_run.state.is_final():
                return

            # update timestamp to not receive logs already shown
            if len(page_logs):
                timestamp = page_logs[-1].timestamp

            await anyio.sleep(poll_interval)

    async def cancel(self):
        """Cancel a flow run by ID."""
        await prefect_utils.cancel(self._prefect_client, self.get_flow_run().id)


class PrefectProgram(infractl.base.Program):
    """Prefect program."""

    flow: prefect.Flow
    deployment_name: str = 'infractl'

    def __init__(
        self,
        path: Union[str, prefect.Flow],
        flow: prefect.Flow,
        name: Optional[str] = None,
    ):
        super().__init__(path=path, name=name)
        self.flow = flow


class PythonProgram(PrefectProgram):
    """Python program wrapped as Prefect flow."""

    program: str = ''
    entrypoint: Optional[str] = None
    files: List[infractl.base.RuntimeFile] = []

    def __init__(self, path: str, name: Optional[str] = None):
        # pylint: disable=import-outside-toplevel, unused-import
        from infractl.prefect import wrapper

        program = load_program(path=wrapper.wrap)
        program.flow.name = prefect_runtime.sanitize(name or pathlib.Path(path).stem)
        super().__init__(path=program.path, flow=program.flow)

        program_path = pathlib.Path(path)
        self.program = program_path.name
        self.entrypoint = name
        self.files = [infractl.base.RuntimeFile(src=str(program_path))]


def load_program(
    path: Union[str, prefect.Flow], name: Optional[str] = None, **kwargs
) -> PrefectProgram:
    """Returns Prefect program."""
    with warnings.catch_warnings():
        # Ignore UserWarning since Prefect complains about loading the same flow definition more
        # than once.
        warnings.simplefilter('ignore', category=UserWarning)
        return _load_program(path, name=name, **kwargs)


def _load_program(
    path: Union[str, prefect.Flow], name: Optional[str] = None, **kwargs
) -> PrefectProgram:
    """Returns Prefect program."""

    flow: Optional[prefect.Flow] = None

    if isinstance(path, str):
        flows = load_flows(path)
        if len(flows) == 0:
            # Use a Prefect flow to wrap a Python program
            return PythonProgram(path=path, name=name)
        elif len(flows) > 1:
            if name is None:
                raise FlowError(f'More than one flow in "{path}", but flow name is not specified')
            flow = flows[name]
        else:
            flow = next(iter(flows.values()))

    if isinstance(path, prefect.Flow):
        flow = path
        # TODO: this does not work with a Prefect flow defined in Jupyter notebook: it does
        # not have __file__ or __module__.
        module_name = getattr(flow, '__module__', None)
        if not module_name:
            raise ValueError('Could not determine flow file location (__module__ not set)')
        module = importlib.import_module(module_name)
        path = getattr(module, '__file__', None)
        if not path:
            raise ValueError('Could not determine flow file location (__file__ not set)')

    return PrefectProgram(path=path, flow=flow, name=flow.name, **kwargs)



"""ICL program."""

from __future__ import annotations

import abc
from typing import Any, Dict, List, Optional, Union


class ProgramRun:
    """Class for representing the results and state of ICL program runs."""

    @abc.abstractmethod
    async def wait(self, poll_interval=5) -> None:
        pass

    @abc.abstractmethod
    def is_scheduled(self) -> bool:
        pass

    @abc.abstractmethod
    def is_pending(self) -> bool:
        pass

    @abc.abstractmethod
    def is_running(self) -> bool:
        pass

    @abc.abstractmethod
    def is_completed(self) -> bool:
        pass

    @abc.abstractmethod
    def is_failed(self) -> bool:
        pass

    @abc.abstractmethod
    def is_crashed(self) -> bool:
        pass

    @abc.abstractmethod
    def is_cancelling(self) -> bool:
        pass

    @abc.abstractmethod
    def is_cancelled(self) -> bool:
        pass

    @abc.abstractmethod
    def is_final(self) -> bool:
        pass

    @abc.abstractmethod
    def is_paused(self) -> bool:
        pass


class Program:
    """Program.

    A Program definition that can be deployed to a specific infrastructure with a specific program
    runtime.
    """

    def __init__(self, path: str, name: Optional[str] = None):
        self.path = path
        self.name = name


class Runnable:
    """Runnable."""

    async def run(
        self,
        parameters: Union[Dict[str, Any], List[str], None] = None,
        timeout: Optional[float] = None,
        detach: bool = False,
    ) -> ProgramRun:
        """Runs this runnable.

        Args:
            parameters: a dictionary of named arguments if a program's entry point is a function,
                a list of arguments otherwise.
            timeout: timeout in seconds to wait for a program completion, `None` (default) to wait
                forever.
            detach: `False` (default) to wait for a program completion, `True` to start the program
                and detach from it.
        """


class DeployedProgram(Runnable):
    """Deployed program."""

    program: Program
    runner: Runnable

    def __init__(self, program: Program, runner: Runnable):
        self.program = program
        self.runner = runner

    async def run(
        self,
        parameters: Union[Dict[str, Any], List[str], None] = None,
        timeout: Optional[float] = None,
        detach: bool = False,
    ) -> ProgramRun:
        """Runs Program."""
        return await self.runner.run(parameters=parameters, timeout=timeout, detach=detach)



"""icl-hub JupyterHub commands."""

import base64
import subprocess
from typing import List, Optional

import click
from kubernetes import client, stream

from infractl.hub import config, kube, root

JUPYTERHUB_NAMESPACE = 'jupyterhub'
ICL_HUB_NAMESPACE = 'icl-hub'


@root.cli.group()
def jupyterhub():
    """Manage JupyterHub."""


@jupyterhub.command('list-sessions')
def list_sessions_cmd():
    """List user sessions and names."""
    for pod in get_user_pods(JUPYTERHUB_NAMESPACE):
        print(
            pod.metadata.name,
            pod.metadata.annotations.get('hub.jupyter.org/username'),
        )


@jupyterhub.command('list-users')
def list_users_cmd():
    """List JupyterHub users."""
    users = list_users()
    if users:
        print('\n'.join(users))


@jupyterhub.command('update-infractl')
def update_infractl_cmd():
    """Update infractl in all JupyterHub sessions."""
    for pod in get_user_pods(JUPYTERHUB_NAMESPACE):
        print('Pod:', pod.metadata.name)
        response = stream.stream(
            kube.api().core_v1().connect_get_namespaced_pod_exec,
            pod.metadata.name,
            JUPYTERHUB_NAMESPACE,
            container='notebook',
            command=['/bin/bash', '-ce', _update_infractl_script()],
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
        )
        print(response)


@jupyterhub.group()
def ssh():
    """Manage SSH access to JupyterHub session."""


@ssh.command('enable')
@click.argument('username')
@click.option('--key', help='Public SSH key')
def enable_ssh_cmd(username: str, key: Optional[str] = None):
    """Enable ssh to JupyterHub session."""
    print(enable_ssh(username, key))


@ssh.command('disable')
@click.argument('username')
def disable_ssh_cmd(username: str):
    """Disable ssh to JupyterHub session."""
    disable_ssh(username)


def get_user_pod(username: str) -> Optional[client.V1Pod]:
    """Returns user pod."""
    for pod in get_user_pods(JUPYTERHUB_NAMESPACE):
        if pod.metadata.annotations.get('hub.jupyter.org/username') == username:
            return pod
    return None


def get_user_pods(namespace: str) -> List[client.V1Pod]:
    """Returns a list of pods with JupyterHub user sessions."""
    response = (
        kube.api()
        .core_v1()
        .list_namespaced_pod(
            namespace=namespace,
            label_selector='component=singleuser-server',
        )
    )
    return response.items


def get_hub_pod(namespace: str) -> client.V1Pod:
    """Returns JupyterHub pod."""
    response = (
        kube.api()
        .core_v1()
        .list_namespaced_pod(
            namespace=namespace,
            label_selector='component=hub',
        )
    )
    return response.items[0]


def get_node_ip(name: str) -> str:
    """Returns node IP address."""
    response = kube.api().core_v1().read_node_status(name)
    for address in response.status.addresses:
        if address.type == 'InternalIP':
            return address.address
    return name


def get_user_id(username: str) -> str:
    """Returns user id."""
    hub_pod = get_hub_pod(JUPYTERHUB_NAMESPACE)
    response = stream.stream(
        kube.api().core_v1().connect_get_namespaced_pod_exec,
        hub_pod.metadata.name,
        JUPYTERHUB_NAMESPACE,
        command=[
            'sqlite3',
            '-readonly',
            'jupyterhub.sqlite',
            f'.param set :username "{username}"',
            'select id from users where name=:username;',
        ],
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )
    return response.strip()


def list_users() -> List[str]:
    """List JupyterHub users."""
    hub_pod = get_hub_pod(JUPYTERHUB_NAMESPACE)
    # TODO: implement a helper method to execute scripts in pod
    response = stream.stream(
        kube.api().core_v1().connect_get_namespaced_pod_exec,
        hub_pod.metadata.name,
        JUPYTERHUB_NAMESPACE,
        command=['sqlite3', '-readonly', 'jupyterhub.sqlite', 'select id, name from users;'],
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
        _preload_content=False,
    )
    response.run_forever()
    if response.returncode != 0:
        raise Exception(response.read_stderr())
    return response.read_stdout().strip().splitlines()


def _update_infractl_script() -> str:
    """Returns Bash script to update infractl."""
    return """
        source /home/jovyan/.conda/etc/profile.d/conda.sh
        conda env list --json | jq -r .envs[] | while read prefix; do
            echo activating conda environment at $prefix
            conda activate $prefix
            if pip show -q infractl 2>/dev/null; then
                echo updating infractl at $prefix
                pip install \
                    --quiet \
                    --upgrade \
                    --extra-index-url http://pypi.glados.intel.com \
                    --trusted-host pypi.glados.intel.com infractl
            fi
        done 
    """


def _enable_ssh_script() -> str:
    """Returns Bash script to enable ssh in JupyterHub session."""
    return """
        if [[ -f /run/sshd.pid && -d /proc/$(cat /run/sshd.pid) ]]; then
            echo sshd is running
        else
            if [[ -x /usr/sbin/sshd ]]; then
                echo sshd is installed
            else
                echo ssh is not installed, trying to install ...
                export DEBIAN_FRONTEND=noninteractive
                sudo -E apt-get update
                sudo -E apt-get install -y --no-install-recommends openssh-server
            fi
            echo starting sshd ...
            sudo mkdir -p -m0755 /var/run/sshd
            sudo /usr/sbin/sshd
            echo sshd is running
        fi
    """


def _add_ssh_key_script(key: str):
    """Adds ssh key."""
    return f"""
        sudo chown jovyan:jovyan /home/jovyan
        mkdir -p ~/.ssh
        echo "{key}" >> ~/.ssh/authorized_keys
        cat ~/.ssh/authorized_keys | sort | uniq > ~/.ssh/authorized_keys.tmp
        mv ~/.ssh/authorized_keys.tmp ~/.ssh/authorized_keys
        chmod 00600 ~/.ssh/authorized_keys
        sudo chmod 00700 /home/jovyan /home/jovyan/.ssh
    """


def _add_private_ssh_key_script(name: str, key: str):
    """Add private ssh key."""
    return f"""
        mkdir -p -m0700 ~/.ssh
        echo "{key}" > ~/.ssh/{name}
        chmod 0600 ~/.ssh/{name}
    """


def _add_jumphost_ssh_key(user_key: str, jumphost_key: str, host: str, port: int = 22):
    """Adds public ssh key to jumphost."""
    subprocess.check_output(
        ['/bin/bash', '-c', _add_private_ssh_key_script('jumphost', jumphost_key)]
    )
    subprocess.check_output(
        [
            '/usr/bin/ssh',
            '-o',
            'UserKnownHostsFile=/dev/null',
            '-o',
            'StrictHostKeyChecking=no',
            '-i',
            '~/.ssh/jumphost',
            '-p',
            port,
            host,
            f'/bin/bash -c {_add_ssh_key_script(user_key)}',
        ]
    )


def enable_ssh(username: str, key: Optional[str] = None) -> str:
    """Enable ssh to JupyterHub session.

    Returns:
        ssh connection string
    """

    # Steps:
    # * Install sshd in the session container and start it in background.
    # * Add public ssh key to ~/.ssh/authorized keys.
    # * Create a Kubernetes NodePort service for ssh port.
    # * Optional: Add public ssh key to the tunnel user.

    # Modes to access NodePort:
    # * Direct
    # * SSH jumphost
    # * SSH proxy

    user_id = get_user_id(username)
    if not user_id:
        raise click.ClickException(f'No such JupyterHub user {username}')
    node_port = int(f'32{user_id.zfill(3)}')

    pod = get_user_pod(username)
    if not pod:
        raise click.ClickException(f'No JupyterHub session for user {username}')

    script = _enable_ssh_script()
    if key:
        script += _add_ssh_key_script(key)

    response = stream.stream(
        kube.api().core_v1().connect_get_namespaced_pod_exec,
        pod.metadata.name,
        JUPYTERHUB_NAMESPACE,
        container='notebook',
        command=['/bin/bash', '-ce', script],
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )
    print(response)

    service_name = f'{pod.metadata.name}-ssh'
    service_body = client.V1Service(
        api_version='v1',
        kind='Service',
        metadata=client.V1ObjectMeta(
            name=service_name,
        ),
        spec=client.V1ServiceSpec(
            type='NodePort',
            selector={
                'hub.jupyter.org/username': pod.metadata.labels['hub.jupyter.org/username'],
                'component': 'singleuser-server',
            },
            ports=[
                client.V1ServicePort(
                    name='ssh',
                    port=22,
                    target_port=22,
                    node_port=node_port,
                    protocol='TCP',
                ),
            ],
        ),
    )

    try:
        kube.api().core_v1().read_namespaced_service(service_name, JUPYTERHUB_NAMESPACE)
    except client.exceptions.ApiException as error:
        if error.status == 404:
            kube.api().core_v1().create_namespaced_service(JUPYTERHUB_NAMESPACE, body=service_body)
        else:
            raise error

    response = kube.api().core_v1().read_namespaced_service(service_name, JUPYTERHUB_NAMESPACE)

    jumphost: str = ''
    if key:
        # if secret ssh-tunnel exists, use it to ssh to the tunnel and add user key
        try:
            response = (
                kube.api().core_v1().read_namespaced_secret('ssh-jumphost', ICL_HUB_NAMESPACE)
            )
            jumphost = base64.b64decode(response.data['host']).decode('utf-8')
            jumphost_key = base64.b64decode(response.data['key']).decode('utf-8')
            _add_jumphost_ssh_key(
                user_key=key,
                jumphost_key=jumphost_key,
                host=jumphost,
            )
        except client.exceptions.ApiException as error:
            if error.status != 404:
                raise error

    ssh_proxy = config.get().get('ingress_domain')
    print('ingress_domain', ssh_proxy)
    ssh_host = ssh_proxy or get_node_ip(pod.spec.node_name)
    ssh_args = f'jovyan@{ssh_host} -p {response.spec.ports[0].node_port}'
    if jumphost:
        return f'ssh -J {jumphost} {ssh_args}'
    else:
        return f'ssh {ssh_args}'


def disable_ssh(username: str):
    """Disable ssh to JupyterHub session."""

    # Steps:
    # * TODO: Stop ssh in session container
    # * Delete service

    pod = get_user_pod(username)
    if not pod:
        raise click.ClickException(f'No JupyterHub session for user {username}')

    name = f'{pod.metadata.name}-ssh'
    try:
        kube.api().core_v1().delete_namespaced_service(name, JUPYTERHUB_NAMESPACE)
    except client.exceptions.ApiException as error:
        if error.status != 404:
            raise error



"""ICL deploy."""


from __future__ import annotations

import asyncio
from typing import Optional

import infractl.api.infrastructure
import infractl.api.runtime
import infractl.base


async def deploy(
    program: infractl.base.Program,
    runtime: Optional[infractl.base.Runtime] = None,
    infrastructure: Optional[infractl.base.Infrastructure] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> infractl.base.DeployedProgram:
    """Deploys a program.

    Args:
        program: a program to deploy, can be a file name, ICL program or Prefect Flow instance.
        runtime: an optional program runtime to use for deployment.
        infrastructure: an optional infrastructure to use for deployment.
        timeout: an optional timeout to use for deployment.
        kwargs: other parameters for deployment.
    """
    if not isinstance(program, infractl.base.Program):
        raise NotImplementedError(f"{program=} don't supported")

    if infrastructure is None:
        infrastructure = infractl.api.infrastructure.default_infrastructure()

    if runtime is None:
        runtime = infractl.api.runtime.default_runtime()

    infrastructure_implementation = infractl.base.get_infrastructure_implementation(infrastructure)
    runtime_implementation = infractl.base.get_runtime_implementation(
        runtime, infrastructure_implementation
    )

    return await asyncio.wait_for(runtime_implementation.deploy(program, **kwargs), timeout=timeout)



"""Wrapper for Prefect engine."""

import argparse
import asyncio
import pathlib
import runpy
import subprocess
import sys
import tempfile


async def run_script(block: str, script: str):
    """Downloads a directory from Prefect storage block and executes a shell script.

    Downloads a directory from the specified Prefect storage block and executes a shell script with
    the specified name in directory. The script is executed in the current directory, it can access
    other files in the temporary directory with:

    ```
    # https://stackoverflow.com/questions/59895/getting-the-source-directory-of-a-bash-script-from-within
    SCRIPT_DIR=$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd )
    ```
    """
    # pylint: disable=import-outside-toplevel
    import prefect.blocks.core as blocks

    print(f'[infractl.prefect.engine] Loading Prefect storage block {block}')
    storage = await blocks.Block.load(block)
    with tempfile.TemporaryDirectory() as dirname:
        await storage.get_directory(local_path=dirname)
        script_path = pathlib.Path(dirname) / script
        if script_path.exists():
            print(f'[infractl.prefect.engine] Running script {script}')
            subprocess.run(
                ['/bin/bash', '-x', script_path.absolute()],
                stdout=sys.stdout,
                stderr=subprocess.STDOUT,
                check=True,
            )
        else:
            print(f'[infractl.prefect.engine] Script {block} not found')


def create_parser() -> argparse.ArgumentParser:
    """Creates argument parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--block', help='Download a directory from this storage block and execute a shell script'
    )
    parser.add_argument(
        '--script',
        default='entrypoint.sh',
        help='Script name to execute (default is "entrypoint.sh")',
    )
    return parser


def main():
    """Entry point."""
    args = create_parser().parse_args()
    if args.block:
        asyncio.run(run_script(args.block, args.script))
    # Delete all command line arguments, prefect.engine does not need them
    sys.argv = sys.argv[0:1]
    runpy.run_module('prefect.engine', run_name='__main__')


if __name__ == '__main__':
    main()



"""ICL runtime."""

from __future__ import annotations

import abc
import os
from typing import Dict, List, Optional

import pydantic

import infractl.base.infrastructure
from infractl.base import registry


class RuntimeDependencies(pydantic.BaseModel):
    """Runtime dependencies.

    Allows declaratively define Python dependencies for different targets,
    such as Prefect, Dask, Ray.
    """

    pip: List[str] = pydantic.Field(
        default_factory=list,
        title='List of pip requirements specifiers',
        description='https://pip.pypa.io/en/stable/reference/requirement-specifiers/',
    )


class RuntimeFile(pydantic.BaseModel):
    """Runtime files.

    Additional files that must exist in the runtime.
    """

    # TODO: support for different schemas, such as 'zip', 's3', and so on
    src: Optional[str] = pydantic.Field(
        default=None,
        title='File source',
        description=(
            'Can be one of: '
            '1. absolute path '
            '2. relative path to the current working directory '
        ),
    )

    dst: Optional[str] = pydantic.Field(
        default=None,
        title='Runtime destination',
        description='Default destination is the runtime current directory',
    )


def parse_dependencies(dependencies: Optional[RuntimeDependencies | Dict]) -> RuntimeDependencies:
    """Parses runtime dependencies."""
    if dependencies is None:
        return RuntimeDependencies()
    if isinstance(dependencies, RuntimeDependencies):
        return dependencies
    if isinstance(dependencies, dict):
        return RuntimeDependencies.parse_obj(dependencies)
    raise NotImplementedError(
        f'{dependencies.__class__.__name__} is not supported as dependency specification'
    )


def parse_files(files: Optional[List[RuntimeFile | Dict | str | os.PathLike]]) -> List[RuntimeFile]:
    """Parses runtime working directory."""
    if files is None:
        return []
    result = []
    for file in files:
        if isinstance(file, str):
            result.append(RuntimeFile(src=file))
        elif isinstance(file, RuntimeFile):
            result.append(file)
        elif isinstance(file, dict):
            # convert os.PathLike to str
            clone = file.copy()
            for key in ('src', 'dst'):
                value = clone.get(key)
                if isinstance(value, os.PathLike):
                    clone[key] = str(value)
            result.append(RuntimeFile.parse_obj(clone))
        elif isinstance(file, os.PathLike):
            result.append(RuntimeFile(src=str(file)))
        else:
            raise NotImplementedError(f'{file.__class__.__name__} is not supported as file item')
    return result


class Runtime:
    """Program runtime specification (for users).

    Allows executing a program in a specific infrastructure.
    """

    def __init__(
        self,
        environment: Optional[Dict[str, str]] = None,
        dependencies: Optional[RuntimeDependencies | Dict] = None,
        files: Optional[List[RuntimeFile | Dict | str]] = None,
        kind: Optional[str] = None,
    ):
        """Creates a new runtime for any compatible infrastructure.

        Args:
            environment: optional dictionary of environment variables
            dependencies: optional runtime dependencies
            files: optional list of runtime files
        """
        self.environment = environment or {}
        self.dependencies = parse_dependencies(dependencies)
        self.files = parse_files(files)
        self.kind = kind


class RuntimeImplementation(metaclass=registry.RegisteredClass):
    """Runtime implementation (for internal usage)."""

    def __init__(
        self,
        runtime: Optional[infractl.base.runtime.Runtime] = None,
        infrastructure_implementation: Optional[
            infractl.base.infrastructure.InfrastructureImplementation
        ] = None,
    ):
        self.runtime = runtime
        self.infrastructure_implementation = infrastructure_implementation

    @abc.abstractmethod
    async def deploy(
        self, program: infractl.base.program.Program, **kwargs
    ) -> infractl.base.DeployedProgram:
        """Deploys a program."""


def get_runtime_implementation(
    runtime: Runtime,
    infrastructure_implementation: infractl.base.infrastructure.InfrastructureImplementation,
) -> RuntimeImplementation:
    runtime_implementation_class = infractl.base.RegisteredClass.get(
        infractl.base.RuntimeImplementation, runtime.kind
    )
    return runtime_implementation_class(runtime, infrastructure_implementation)



#!/usr/bin/env python3

"""icl-hub CLI."""

# pylint: disable=unused-import
import infractl.hub.config
import infractl.hub.jupyterhub
import infractl.hub.server
from infractl.hub.root import cli

if __name__ == '__main__':
    cli()



"""Utils for Prefect runtime."""

from prefect.client.schemas.objects import StateType
from prefect.client.schemas.responses import SetStateStatus
from prefect.exceptions import ObjectNotFound
from prefect.states import State

from infractl.logging import get_logger

logger = get_logger()


async def cancel(prefect_client, flow_id):
    """Cancel a flow run by ID."""
    # Copied (with modification) from:
    #  @flow_run_app.command()
    #  async def cancel(id: UUID):
    cancelling_state = State(type=StateType.CANCELLING)
    try:
        result = await prefect_client.set_flow_run_state(
            flow_run_id=flow_id, state=cancelling_state
        )
    except ObjectNotFound as exc:
        raise RuntimeError(f"Flow run '{flow_id}' not found!") from exc

    if result.status == SetStateStatus.ABORT:
        raise RuntimeError(
            f"Flow run '{flow_id}' was unable to be cancelled. "
            + f"Reason: '{result.details.reason}'"
        )

    logger.info("Flow run with id: '%s' was successfully scheduled for cancellation.", flow_id)



"""Prefect specific module."""
import re

from infractl.plugins.prefect_runtime.program import (
    PrefectProgram,
    PrefectProgramRun,
    PythonProgram,
    load_program,
)
from infractl.plugins.prefect_runtime.runtime import PrefectRuntimeImplementation


def sanitize(identity: str) -> str:
    """Sanitizes the value to make it compatible with Prefect names."""
    # replace non-alphanumeric characters with '-'
    identity = re.sub('[^0-9a-zA-Z]+', '-', identity.lower())
    # replace repeating '-' with a single one, strip heading and leading '-'
    return re.sub('--+', '-', identity).strip('-')


__all__ = [
    'PrefectRuntimeImplementation',
    'PrefectProgram',
    'PrefectProgramRun',
    'PythonProgram',
    'load_program',
    'sanitize',
]



"""ICL infrastructure."""

from __future__ import annotations

from typing import Optional, Tuple, Union

import pydantic

from infractl.base import registry


class Infrastructure(pydantic.BaseModel, extra=pydantic.Extra.allow):
    """Infrastructure specification (for users)."""

    kind: str = pydantic.Field(
        title='Infrastructure kind.',
    )

    address: str = pydantic.Field(
        title='Address of cluster, empty string for the current one',
        default='',
    )

    gpus: Optional[Union[int, Tuple[str, int]]] = pydantic.Field(
        title='Request a number of any GPUs or a specific GPUs.',
        default=None,
    )

    @pydantic.validator('gpus')
    @classmethod
    def _check_gpus_tuple(cls, gpus):
        if isinstance(gpus, int):
            gpus = ('gpu', gpus)
        return gpus


class InfrastructureImplementation(metaclass=registry.RegisteredClass):
    """Infrastructure implementation (for internal usage)."""

    def __init__(
        self,
        infrastructure: Infrastructure,
    ):
        self.infrastructure = infrastructure

    @property
    def address(self) -> str:
        return self.infrastructure.address


def get_infrastructure_implementation(
    infrastructure: Infrastructure,
) -> InfrastructureImplementation:
    infrastructure_implementation_class = registry.RegisteredClass.get(
        InfrastructureImplementation, infrastructure.kind
    )
    return infrastructure_implementation_class(infrastructure)



"""Register and discover classes in runtime."""


from typing import Any, Dict, Optional, Tuple, Type


# pylint: disable=protected-access, unused-argument
class RegisteredClass(type):
    """A metaclass to make classes findable through a registry.

    RegisteredClass is a metaclass that lets you define classes whose subclasses you can look up
    by name. This is useful if (for example) you want to define an abstract class in a library,
    have users of the library define their own implementations of that class, and pick which
    implementation you want to use at runtime based on a parameter. For example::

       class AbstractWorker(metaclass=RegisteredClass):
           ... define some stuff ...

           @classmethod
           def make(cls, worker: str, other_arguments) -> 'AbstractWorker':
               return RegisteredClass.get(AbstractWorker, worker)(other_arguments)

       # In another file
       class RealWorker(AbstractWorker):
           ... just a normal subclass ...

    You access this registration using two static methods:

    * ``RegisteredClass.get(superclass, name)`` returns a named subclass of the superclass
    * ``RegisteredClass.subclasses(superclass)`` returns all registered subclasses of that class.

    Or via class methods:

    * ``AbstractWorker.get_subclass(name)  # type: ignore``
    * ``AbstractWorker.subclasses()  # type: ignore``

    You can also define "intermediate" classes which don't themselves appear in the registry. For
    example::

        class RemoteWorker(AbstractWorker, register=False):  # type: ignore
            ... some partial implementation of AbstractWorker ...

        class RealRemoteWorker(RemoteWorker):
            ... a concrete worker ...

    Then ``RegisteredClass.subclasses(AbstractWorker)`` will return `RealWorker` and
    `RealRemoteWorker` (but not `RemoteWorker`), and ``RegisteredClass.subclasses(RemoteWorker)``
    will return `RealRemoteWorker`.

    Every subclass must be registered with a unique name, which by default is just the name of the
    class. You can override this with ``registration_name='Foo'``.


    MYPY WARNING: There are some bugs in the way mypy handles dynamic type declarations. As a
    result, if you use any of the class methods (rather than ``RegistrationClass.*``) or if you
    pass any arguments like `register` or `registration_name`, you have to mark the line as
    ``# type: ignore``.  Sorry.
    """

    # Type signature to make mypy happy: All types that use this as a metaclass will have this
    # as a class variable.
    _registry: Dict[str, 'RegisteredClass']

    @staticmethod
    def get(superclass: 'RegisteredClass', name: str) -> 'RegisteredClass':
        """Get a named subclass of superclass.

        Args:
          superclass: A class whose metaclass is RegisteredClass.
          name: The name under which a subclass was registered -- usually the name of the class,
            unless you've set registration_name in its class declaration.
        """
        if name not in superclass._registry:
            raise KeyError(f'No subclass "{name}" of "{superclass.__name__}" has been registered.')
        result = superclass._registry[name]
        return result

    @staticmethod
    def subclasses(superclass: 'RegisteredClass') -> Dict[str, 'RegisteredClass']:
        """Return all the subclasses of a given registered class."""
        if not hasattr(superclass, '_registry'):
            raise TypeError(f'The class {superclass} is not a registered class.')
        # This comprehension serves two purposes. First, it means we don't hand the caller a mutable
        # pointer into the real registry, which could cause all hell to break loose. Second, it
        # means that if the caller passes a *subclass* of a registered class -- e.g., a partial
        # implementation -- we correctly return its subclasses!
        return {
            name: class_
            for name, class_ in superclass._registry.items()
            if issubclass(class_, superclass) and class_ != superclass
        }

    def __new__(
        mcs,
        name: str,
        bases: Tuple[Type],
        namespace: Dict[str, Any],
        register: bool = True,
        registration_name: Optional[str] = None,
    ) -> Any:
        # This function gets called when a new registered class (ie one whose metaclass is
        # RegisteredClass) is declared. We set up its registry and its subclass initializer.
        class_ = type(name, bases, namespace)

        def init_subclass(
            cls: type, register: bool = True, registration_name: Optional[str] = None
        ) -> None:
            if register:
                registration_name = registration_name or cls.__name__
                # Don't register the root class itself, that's pretty much never useful.
                if registration_name == name:
                    return
                if registration_name in cls._registry:  # type: ignore
                    raise ValueError(
                        f'There is already a registered subclass of {name} named'
                        f' {registration_name}'
                    )
                cls._registry[registration_name] = cls  # type: ignore

        # Doing it this way makes mypy happier.
        registry: Dict[str, Type[class_]] = {}  # type: ignore
        setattr(class_, '_registry', registry)
        setattr(class_, '__init_subclass__', classmethod(init_subclass))

        # Also define some class implementations of the static functions. Alas, using these makes
        # mypy freak out, so if you do use them, you'll have to mark it # type: ignore.

        def get_subclass(superclass: 'RegisteredClass', name: str) -> 'RegisteredClass':
            return RegisteredClass.get(superclass, name)

        def subclasses(superclass: 'RegisteredClass') -> Dict[str, 'RegisteredClass']:
            return RegisteredClass.subclasses(superclass)

        setattr(class_, 'get_subclass', classmethod(get_subclass))
        setattr(class_, 'subclasses', classmethod(subclasses))

        return class_



"""ICL settings."""

import dynaconf

SETTINGS = dynaconf.Dynaconf(
    envvar_prefix='X1',  # export envvars with `export X1_FOO=bar`
    settings_files=['/etc/x1/settings.yaml', '.x1/settings.yaml'],
    merge_enabled=True,
)



"""Builds and pushes a Docker image to Docker registry using ICL cluster.

1. Upload the specified context recursively (with .dockerignore) to the cluster storage.
2. Create a Kubernetes Job that
    a. downloads the context to a temporary directory
    b. builds a Docker image
    c. pushes the Docker image to the container registry

The code needs to set the job TTL, so it will be deleted after completion.
"""
import copy
import logging
import re
import warnings
from typing import Any, Dict, Optional

import infractl
import infractl.base
import infractl.docker
import infractl.logging


class Builder(infractl.docker.Builder):
    """Builds and pushes a Docker image to Docker registry using ICL cluster."""

    logger: logging.Logger

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(
        self,
        stream_callback: Optional[infractl.docker.StreamCallback] = infractl.docker.stdout_callback,
        **kwargs,
    ) -> infractl.docker.Image:
        """Builds and pushes a Docker image to Docker registry using ICL cluster.

        Arguments:
        https://docker-py.readthedocs.io/en/stable/images.html#docker.models.images.ImageCollection.build
        """

        # TODO: move to the global import after fixing the issue with logger
        # pylint: disable=import-outside-toplevel, unused-import
        from prefect import futures

        # TODO: image full tag (name:tag) is currently required!
        for key in ('path', 'tag'):
            if key not in kwargs:
                raise ValueError(f'Argument {key} is required')

        path = kwargs['path']
        if not path.endswith('/'):
            path = f'{path}/'

        build_args = kwargs.copy()
        build_args['path'] = path

        with warnings.catch_warnings():
            # Disabling warning in JupyterLub:
            # UserWarning: `sync` called from an asynchronous context; you should `await` the async
            # function directly instead.
            warnings.simplefilter('ignore', category=UserWarning)
            futures.sync(self._async_build, **build_args)

        return infractl.docker.Image.from_full_name(kwargs['tag'])

    async def _async_build(self, **build_args):
        """Deploys and runs Prefect builder."""
        self.logger = infractl.logging.get_logger(__name__)

        # pylint: disable=import-outside-toplevel, unused-import
        from infractl.docker.prefect import builder

        runtime = infractl.runtime(
            # TODO: consider using a custom image for faster start
            dependencies={
                'pip': ['psutil'],
            },
            files=[
                {'src': build_args['path'], 'dst': 'context/'},
            ],
        )

        self.logger.info('Deploying Prefect builder flow')
        program = await infractl.deploy(
            infractl.program(builder.build),
            runtime=runtime,
            infrastructure=self.infrastructure.infrastructure,
            manifest_filter=self.manifest_filter,
        )

        self.logger.info('Running Prefect builder flow')
        flow_run = await program.run(
            parameters={
                'build_args': build_args,
                'registry': self.registry_internal_endpoint,
            },
            detach=True,
        )

        self.logger.info('Prefect flow run: %s', flow_run)
        await flow_run.stream_logs()
        if not flow_run.is_completed():
            raise infractl.docker.BuilderError(f'Remote build failed in {flow_run}')

    def manifest_filter(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        """Filters Kubernetes Job manifest."""
        # TODO: use registry mirror
        registry_mirror = None
        insecure_registries = [re.sub(r'^https?://', '', self.registry_internal_endpoint)]

        args = []
        if registry_mirror:
            args.extend(['--registry-mirror', registry_mirror])
        if insecure_registries:
            for registry in insecure_registries:
                args.extend(['--insecure-registry', registry])

        manifest = copy.deepcopy(manifest)

        job = manifest['spec']
        pod = job['template']['spec']
        prefect_container = pod['containers'][0]

        # pod volumes
        volumes = pod.setdefault('volumes', [])
        volumes.append({'name': 'certs-client', 'emptyDir': {}})

        # dind container
        pod['containers'].append(
            {
                'name': 'dind',
                'image': 'docker:dind',
                'imagePullPolicy': 'IfNotPresent',
                'env': [
                    {'name': 'DOCKER_TLS_CERTDIR', 'value': '/certs'},
                ],
                'args': args,
                'volumeMounts': [
                    {'name': 'certs-client', 'mountPath': '/certs/client'},
                ],
                'securityContext': {
                    'privileged': True,
                },
            },
        )

        # set TTL for the job to be automatically removed after 10 minutes
        job['ttlSecondsAfterFinished'] = 600

        # run dind and prefect in the same process namespace, so prefect can see dind process and
        # terminate it when necessary
        pod['shareProcessNamespace'] = True

        # set environment variables in prefect-job container
        prefect_env = prefect_container.setdefault('env', {})
        prefect_env.append({'name': 'DOCKER_HOST', 'value': 'tcp://localhost:2376'})
        prefect_env.append({'name': 'DOCKER_TLS_VERIFY', 'value': '1'})
        prefect_env.append({'name': 'DOCKER_CERT_PATH', 'value': '/certs/client'})

        # attach docker certs to prefect container
        volume_mounts = prefect_container.setdefault('volumeMounts', [])
        volume_mounts.append({'name': 'certs-client', 'mountPath': '/certs/client'})

        # run prefect container as privileged to allow killing docker process
        prefect_container['securityContext'] = {'privileged': True}

        return manifest



#!/usr/bin/env python3

"""ICL CLI."""

import logging
import os
import urllib.parse
from typing import Optional

import click
import requests

import infractl.base


@click.group()
def main():
    """ICL CLI tool."""
    logging.basicConfig(level=logging.INFO)


@main.group()
def ssh():
    """Manage SSH access to JupyterHub session."""
    if 'JUPYTERHUB_USER' not in os.environ:
        raise click.ClickException(
            'Environment variable JUPYTERHUB_USER not set, are you in JupyterHub session?'
        )


@ssh.command('enable')
@click.option('--key', help='Public SSH key')
def enable_ssh_cmd(key: Optional[str] = None):
    """Enable ssh to JupyterHub session."""
    print('Enabling SSH access ...')
    print(
        'Use the following command to log in to your session:',
        enable_ssh(os.getenv('JUPYTERHUB_USER'), key),
    )


@ssh.command('disable')
def disable_ssh_cmd():
    """Disable ssh to JupyterHub session."""
    print('Disabling SSH access ...')
    print(disable_ssh(os.getenv('JUPYTERHUB_USER')))


def get_hub_url() -> str:
    """Returns icl-hub address."""
    # TODO: support HTTP and HTTPS schemas, move to upper level
    return infractl.base.SETTINGS.get(
        'hub_url',
        f'http://hub.{infractl.base.SETTINGS.get("default_address", "localtest.me")}',
    )


def enable_ssh(username: str, key: Optional[str] = None) -> str:
    """Enable ssh to JupyterHub session.

    Returns:
        ssh connection string
    """
    url = f'{get_hub_url()}/jupyterhub/ssh/enable/{urllib.parse.quote(username)}'
    response = requests.post(url, json={'key': key}, timeout=300)
    response.raise_for_status()
    return str(response.json())


def disable_ssh(username: str) -> str:
    """Disable ssh to JupyterHub session."""
    url = f'{get_hub_url()}/jupyterhub/ssh/disable/{urllib.parse.quote(username)}'
    response = requests.post(url, timeout=60)
    response.raise_for_status()
    return str(response.json())


if __name__ == '__main__':
    main()



"""ICL program."""

import infractl.base


def program(*args, kind: str = 'prefect', **kwargs) -> infractl.base.Program:
    """Create a program.

    Args:
        args:
            Any positional arguments for `kind` program.
        kind:
            Kind of program, default is "prefect".
        kwargs:
            Any keywords arguments for `kind` program.

    Examples:
        1. Python file with Prefect flow:
            program('my_flow.py')
        2. Imported Prefect flow:
            from my_flow import flow1
            program(flow1)
    """
    if kind == 'prefect':
        # pylint: disable=import-outside-toplevel
        from infractl.plugins import prefect_runtime

        return prefect_runtime.load_program(*args, **kwargs)

    raise NotImplementedError(f'Program kind f{kind} is not supported')



"""ICL runtime."""

import infractl.base


def default_runtime() -> infractl.base.Runtime:
    """Returns a default program runtime."""
    return infractl.base.Runtime(kind='prefect')


def runtime(*args, kind: str = 'prefect', **kwargs) -> infractl.base.Runtime:
    if kind == 'prefect':
        # pylint: disable=import-outside-toplevel, unused-import, redefined-outer-name
        import infractl.plugins.prefect_runtime.runtime  # noqa

        return infractl.base.Runtime(*args, kind=kind, **kwargs)

    raise NotImplementedError(f'Runtime {kind} is not implemented')



"""icl-hub configuration."""

import pathlib
import tempfile
from typing import Any, Dict, Optional

import dynaconf
import dynaconf.loaders

from infractl.hub import root

_CONFIG: Optional[dynaconf.Dynaconf] = None


@root.cli.group()
def config():
    """Manage ICL hub configuration."""


@config.command()
def view():
    """View configuration."""
    with tempfile.TemporaryDirectory() as temp_dir:
        config_path = str(pathlib.Path(temp_dir) / 'hub.toml')
        dynaconf.loaders.write(config_path, get_dict())
        with open(config_path, mode='r', encoding='utf-8') as config_file:
            print(config_file.read())


def get() -> dynaconf.Dynaconf:
    """Gets configuration."""
    global _CONFIG
    if _CONFIG is None:
        _CONFIG = dynaconf.Dynaconf(
            envvar_prefix='X1',
            # override by SETTINGS_FILE_FOR_DYNACONF or SETTINGS_FILES_FOR_DYNACONF
            preload=['.x1/hub.toml'],
            merge_enabled=True,
        )
    return _CONFIG


def get_dict() -> Dict[str, Any]:
    """Gets configuration as dictionary.

    Used to return all configuration keys and values.
    """
    # Make top level keys lower case since dynaconf makes them uppercase by default
    return {key.lower(): value for key, value in get().as_dict().items()}



"""ICL Docker API.

This API allows building custom Docker images and pushing them to a Docker registry.

* It leverages the existing local Docker, if exists.
* If Docker cannot be used, it uses the existing ICL cluster to build an image.
* It can be used inside (for example, in JupyterLab) and outside an ICL cluster (for example, from
  user's laptop).
* It leverages the existing infrastructure to access ICL cluster and its resources.
* It supports direct K8s cluster API and ICL cluster API, when the former is not available.
"""

from __future__ import annotations

import enum
import os
import re
import shutil
from typing import Any, Callable, Optional, Union

import pydantic
import python_docker.registry
import requests.exceptions

import infractl.api
import infractl.api.infrastructure
import infractl.base

StreamCallback = Callable[[str], None]


def stdout_callback(line: str):
    """Stream callback that prints all messages to stdout."""
    print(line)


class BuilderError(Exception):
    """Base class for builder errors."""


class BuilderKind(enum.Enum):
    """Kind of Docker builder."""

    AUTO = 1
    """Auto detect which Docker builder to use (default)."""

    DOCKER = 2
    """Use local Docker daemon."""

    PREFECT = 3
    """Use Prefect installed in the provided infrastructure."""

    KUBERNETES = 4
    """Use Kubernetes cluster in the provided infrastructure."""

    @classmethod
    def list(cls):
        """Returns a list of possible kinds."""
        return list(map(lambda c: c.name, cls))


class Image(pydantic.BaseModel):
    """Docker image."""

    id: Optional[str] = None
    registry: Optional[str] = None
    name: Optional[str] = None
    tag: Optional[str] = None

    @classmethod
    # pylint: disable=redefined-builtin
    def from_full_name(cls, full_name: Optional[str] = None, id: Optional[str] = None):
        """Returns Image instance for a full name.

        Docker terminology for images is quite confusing: their "tag" means a "full" tag such as
        "ubuntu:22.04" and a "real" tag such as "22.04". This function accepts a full tag and
        optional id and returns an Image object.

        Examples:
        'ubuntu' -> name='ubuntu'
        'ubuntu:22.04' -> name='ubuntu', tag='22.04'
        'docker.io/ubuntu' -> registry='docker.io', name='ubuntu'
        'docker.io/ubuntu:22.04' -> registry='docker.io', name='ubuntu', tag='22.04'
        'localhost:5000/ubuntu' -> registry='localhost:5000', name='ubuntu'
        'localhost:5000/ubuntu:22.04' -> registry='localhost:5000', name='ubuntu', tag='22.04'
        """
        registry: Optional[str] = None
        name: Optional[str] = None
        tag: Optional[str] = None

        if full_name:
            parts = full_name.rsplit('/', 1)
            if len(parts) == 2:
                registry = parts[0]
                full_name = parts[1]

            parts = full_name.rsplit(':', 1)
            if len(parts) == 1:
                name = full_name
            elif len(parts) == 2:
                name = parts[0]
                tag = parts[1]

        return Image(id=id, registry=registry, name=name, tag=tag)

    @property
    def full_name(self):
        """Returns a full name for the image, or its id."""
        if not self.name:
            return self.id
        registry = f'{self.registry}/' if self.registry else ''
        tag = f':{self.tag}' if self.tag else ''
        return f'{registry}{self.name}{tag}'


class Builder:
    """Builds and pushes a Docker image to Docker registry."""

    infrastructure: infractl.base.InfrastructureImplementation
    registry: Optional[str] = None

    def __init__(
        self,
        infrastructure: Optional[infractl.base.Infrastructure] = None,
        registry: Optional[str] = None,
    ):
        """Docker builder.

        Args:
            infrastructure: ICL infrastructure, default one is used when not specified.
            registry: Docker registry, default one from infrastructure is used when not specified.
        """
        self.infrastructure = infractl.base.get_infrastructure_implementation(
            infrastructure or infractl.api.infrastructure.default_infrastructure()
        )
        self.registry = registry

    def build(self, stream_callback: Optional[StreamCallback] = stdout_callback) -> Image:
        """Builds and pushes a Docker image to Docker registry."""

    def image_exists(self, tag: str) -> bool:
        """Checks if the image exists in the registry."""
        image = infractl.docker.Image.from_full_name(full_name=tag)
        pd_registry = python_docker.registry.Registry(hostname=self.registry_endpoint)
        try:
            pd_registry.get_manifest(image=image.name, tag=image.tag)
        except requests.exceptions.HTTPError as error:
            if error.response is not None and error.response.status_code == 404:
                return False
            raise error
        return True

    @property
    def registry_endpoint(self):
        """Registry endpoint."""
        if self.registry:
            if re.search(r'^https?://', self.registry):
                return self.registry
            # Assume that user-specified registry is HTTPS
            return f'https://{self.registry}'
        # Try to detect if running in Kubernetes
        if 'KUBERNETES_SERVICE_HOST' in os.environ:
            return self.registry_internal_endpoint
        else:
            return self.registry_external_endpoint

    @property
    def registry_internal_endpoint(self):
        """Registry internal endpoint."""
        return self.settings(
            'registry_internal_endpoint',
            'http://docker-registry.docker-registry.svc.cluster.local:5000',
        )

    @property
    def registry_external_endpoint(self):
        """Registry internal endpoint."""
        return self.settings(
            'registry_external_endpoint',
            f'http://registry.{self.infrastructure.address}',
        )

    def settings(self, name: str, default_value: Any = None) -> Any:
        """Return a setting value for this infrastructure address."""
        # TODO: this needs to be supported by settings (see also HCL configuration)
        return infractl.base.SETTINGS.get(f'{self.infrastructure.address}.{name}', default_value)


def builder(
    infrastructure: Optional[infractl.base.Infrastructure] = None,
    registry: Optional[str] = None,
    kind: Union[str | BuilderKind] = BuilderKind.AUTO,
):
    """Returns a new Docker image builder.

    Args:
        infrastructure: ICL infrastructure, default one is used when not specified.
        registry: Docker registry, default one from infrastructure is used when not specified.
        kind: kind of Docker builder.
    """
    if isinstance(kind, str):
        kinds = BuilderKind.list()
        if kind.upper() not in kinds:
            raise NotImplementedError(
                f'{kind=} is not supported, possible kinds are {", ".join(kinds).lower()}'
            )
        kind = BuilderKind[kind.upper()]

    if kind == BuilderKind.AUTO:
        if shutil.which('docker'):
            kind = BuilderKind.DOCKER
        else:
            kind = BuilderKind.PREFECT

    if kind == BuilderKind.DOCKER:
        # pylint: disable=import-outside-toplevel, unused-import
        from infractl.docker import local

        return local.Builder(infrastructure=infrastructure, registry=registry)

    if kind == BuilderKind.PREFECT:
        # pylint: disable=import-outside-toplevel, unused-import
        from infractl.docker import remote

        return remote.Builder(infrastructure=infrastructure, registry=registry)

    raise NotImplementedError(f'{kind=} is not supported')
